import { PartialTweet } from "./TwitterTypes";
export declare const TweetSearcher: {
    /**
     * Keywords to use in search query.
     * Each keyword represents a pair **keyword**:**content**.
     *
     * Keywords are used to add search parameters (conditions) that requires a user input
     * (like tweets before a specific date, retweets of a desired user...).
     *
     * If you want to add a condition that is static (test if is a retweet,
     * test if tweet has medias...), see `.static_validators`.
     *
     * See how to add keywords in `TweetSearchValidator` interface.
     */
    validators: TweetSearchValidator[];
    /**
     * Defined "static validators": Validators that does not depends on user query,
     * like test if a tweet is a retweet or keep only tweets with medias.
     *
     * Validators are represented by its name (in key)
     * and the validator itself (the check function, in value).
     *
     * You can use them when you search tweets by specifing the static validators names
     * in the `static_validators` parameter of `search()` method.
     */
    static_validators: {
        [staticName: string]: ValidatorExecFunction;
    };
    /**
     * Search into **tweets** using **query**.
     *
     * Query can contain keywords.
     * Default available keywords are:
     * - `since:{YYYY-MM-DD}`
     * - `until:{YYYY-MM-DD}`
     * - `from:{screen_name,screen_name_2,screen_name_3}`
     * - `retweet_of:{screen_name} (use of , is allowed, like from:)`
     *
     * To be in the result array, a tweet must validate ALL the keywords.
     *
     * You can add your custom validators by pushing
     * to **this.validators** a object implementing `TweetSearchValidator`.
     *
     * @param tweets Partial tweets array.
     *
     * @param query String to be searched for in the tweet.text / screen_name.
     * You should remember that in tweets, `>` and `<` are encoded as `&gt;` and `&lt;`.
     *
     * @param is_regex If the string should be considered as a regex during text.match or not.
     * - If **is_regex** is `true`, regex will be enabled without flags.
     * - If **is_regex** is `false`, **query** will be regex escaped.
     * - If **is_regex** is a `String`, regex will be enabled with **is_regex** as flags.
     *
     * Validators inside the query, such as `since:2018-01-01` will be removed from the query regex/string.
     *
     * @param static_validators Array of static validators names that should used.
     * See `.static_validators`.
     *
     * Default defined static validators are:
     * - `retweets_only`
     * - `medias_only`
     * - `videos_only`
     * - `no_retweets`
     *
     * @param search_in Tweet properties to search. This is NOT dynamic you can't specify the property you want.
     * Available properties are:
     * - `text`
     * - `user.screen_name`
     * - `user.name`
     *
     * Default activated properties are `text` and `user.screen_name`.
     *
     * @throws Format `{keyword}: Invalid query` if user input is invalid
     * for a specific validator.
     *
     * @throws `Validator {name} does not exists` when a invalid static validator is used.
     */
    search(tweets: PartialTweet[], query: string, is_regex?: string | boolean, static_validators?: string[], search_in?: string[]): PartialTweet[];
    /** Reset validators. This does NOT work if you have modified the inner initial objects ! */
    reset(): void;
};
export default TweetSearcher;
export interface TweetSearchValidator {
    /**
     * Keyword (word used to identify you keyword, before the **:**.
     * Will be interpoled into search regex.
     * Should **NOT** contain parenthesis !
     */
    keyword: string;
    /**
     * Method that generate a validator function with user input in your **query**.
     * This method **should** return a *function*.
     *
     * You should verify user-input stored inside {user_query}.
     *
     * If the return type is null-ish (`undefined`, `null`, `false`,...), search will be **aborted**
     * with thrown **unwell-formed query error**.
     *
     * ---
     *
     * For example, if you want to validate tweets that are above a date
     * entered by the user, with the schema: `keyword:YYYY-MM-DD`,
     * `YYYY-MM-DD` entered by the user will be given in parameter to this attribute (`.validator`),
     * which should return a function that test if the tweet date is above user entered `YYYY-MM-DD`.
     *
     * **EXAMPLE**
     *
     * *Do not use this example in real case, it does not check well user input.*
     *
     * ```ts
     * validator: user_query => {
     *  // Parse user input
     *  const time = (new Date(user_query)).getTime();
     *
     *  // Check some data validity
     *  if (isNaN(time)) {
     *    // Data could not be used, validator is invalid, returns undefined.
     *    // This will throw an Error.
     *    return undefined;
     *  }
     *
     *  // Return the validator function that returns true if tweet date >= user's date.
     *  return tweet => dateFromTweet(tweet).getTime() >= time;
     * }
     * ```
     */
    validator: (user_query: string) => ValidatorExecFunction;
}
declare type ValidatorExecFunction = (tweet: PartialTweet) => boolean;
export declare type TweetSearchStaticValidator = ValidatorExecFunction;
